"""
Result
======

The pypesto.Result object contains all results generated by
the pypesto components. It contains sub-results for
optimization, profiling, sampling.
"""

from .optimize import OptimizeResult, OptimizerResult
from .predict import PredictionConditionResult, PredictionResult
from .profile import ProfileResult, ProfilerResult
from .result import Result
from .sample import McmcPtResult, SampleResult
